/*
 GBxCart RW - GBA 128MB Multi-game Flasher v1.0
 Version: 1.0a
 Author: Alex from insideGadgets (www.insidegadgets.com)
 Created: 19/06/2021
 Last Modified: 14/08/2021
 License: CC-BY-NC-SA
 
 This program allows you to write multiple games to our GBA 128MB Multi-game flash cart
 
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <time.h>
#include <math.h>

#ifdef _WIN32
#include <windows.h>
#else
#include <unistd.h>
#endif

#include "setup.h" // See defines, variables, constants, functions here

int main(int argc, char **argv) {
	
	printf("GBxCart RW - GBA 128MB Multi-Game Flasher v1.0a by insideGadgets\n");
	printf("################################################################\n");
	printf("Run this program by itself to change into a game, change a game's FRAM bank or rename a game\n\n");
	
	// Check arguments
	if (argc >= 2) {
		read_config();
		
		// Open COM port
		if (com_test_port() == 0) {
			printf("Device not connected and couldn't be auto detected. Please make sure the COM port isn't open elsewhere. If you are using Linux/Mac, make sure you use \"sudo\".\n");
			read_one_letter();
			return 1;
		}
		printf("Connected on COM port: %i at ", cport_nr+1);
		
		// Break out of any existing functions on ATmega
		set_mode('0');
		RS232_flushRX(cport_nr);
		
		// Get cartridge mode - Gameboy or Gameboy Advance
		cartridgeMode = request_value(CART_MODE);
		
		// Get PCB version
		gbxcartPcbVersion = request_value(READ_PCB_VERSION);
		xmas_wake_up();
		
		// Increase to 1.7Mbit baud for v1.4 PCB
		#ifdef _WIN32
		if (gbxcartPcbVersion == PCB_1_4) {
			set_mode(USART_1_7M_SPEED);
			RS232_CloseComport(cport_nr);
			delay_ms(200);
			
			if (RS232_OpenComport(cport_nr, 1700000, "8N1") == 0) {
				printf("1.7M Baud\n");
			}
		}
		else {
			printf("1M Baud\n");
		}
		#else
			printf("1M Baud\n");
		#endif
		
		char pcbVersionString[20];
		if (gbxcartPcbVersion == PCB_1_0) {
			strncpy(pcbVersionString, "v1.0", 5);
		}
		else if (gbxcartPcbVersion == PCB_1_1) {
			strncpy(pcbVersionString, "v1.1/v1.2", 10);
		}
		else if (gbxcartPcbVersion == PCB_1_3) {
			strncpy(pcbVersionString, "v1.3", 5);
		}
		else if (gbxcartPcbVersion == PCB_1_4) {
			strncpy(pcbVersionString, "v1.4", 5);
		}
		else if (gbxcartPcbVersion == MINI) {
			strncpy(pcbVersionString, "Mini", 5);
		}
		else if (gbxcartPcbVersion == GBXMAS) {
			strncpy(pcbVersionString, "XMAS", 5);
		}
		
		// Get firmware version
		gbxcartFirmwareVersion = request_value(READ_FIRMWARE_VERSION);
		printf("PCB version: %s, Firmware version: %i\n", pcbVersionString, gbxcartFirmwareVersion);
		
		
		// Power up to GBA mode
		set_mode(VOLTAGE_3_3V);
		
		// GBx v1.4 - Power up the cart if not already powered up and flush buffer
		gbx_cart_power_up();
		RS232_flushRX(cport_nr);
		
		if (gbxcartPcbVersion == GBXMAS) {
			set_mode('!');
			delay_ms(50); // Wait for ATmega169 to WDT reset if more than 2 mins idle
		}
		
		// Get cartridge mode - Gameboy or Gameboy Advance
		cartridgeMode = request_value(CART_MODE);
		
		
		if (gbxcartPcbVersion == PCB_1_0) {
			printf("\nPCB v1.0 is not supported for this function.");
			read_one_letter();
			return 1;
		}
		
		if (gbxcartFirmwareVersion <= 8) {
			printf("Firmware R9 or higher is required for this functionality.\n");
			read_one_letter();
			return 1;
		}
		
		#ifdef TIMETEST
			// Timing test
			printf("Timing started\n");
			clock_t t;
			t = clock();
		#endif
		
		// Separate the file name
		char filenameOnly[100];
		char filename[254];
		strncpy(filename, argv[1], 253);
		
#ifdef _WIN32
		char *token = strtok(filename, "\\");
		while (token != NULL) {
			strncpy(filenameOnly, token, 99);
			token = strtok(NULL, "\\");
		}
#else
		char *token = strtok(filename, "/");
		while (token != NULL) {
			strncpy(filenameOnly, token, 99);
			token = strtok(NULL, "/");
		}
#endif
		
		uint32_t readBytes = 0;
		uint16_t sector = 0;
		long fileSize = 0;
		
		// Grab file size
		FILE *romFile = fopen(argv[1], "rb");
		if (romFile != NULL) {
			fseek(romFile, 0, SEEK_END);
			fileSize = ftell(romFile);
			fseek(romFile, 0, SEEK_SET);
		}
		else {
			printf("\n%s \nFile not found\n", argv[1]);
			read_one_letter();
			return 1;
		}
		
		// Find last 0xFF in the file
		long lastTrimLocation = fileSize;
		long originalFileSize = fileSize;
		uint8_t endTrim = 0;
		uint8_t buffer[64];		
		for (long x = fileSize-64; x > 0; x -= 64) {
			fseek(romFile, x, SEEK_SET);
			fread(buffer, 1, 64, romFile);
			
			for (int8_t c = 63; c >= 0; c--) {
				if (buffer[c] == 0xFF) {
					lastTrimLocation--;
				}
				else {
					endTrim = 1;
					if ((lastTrimLocation+64) < fileSize) {
						while (lastTrimLocation % 64 != 0) { // Align to 64 bytes
							lastTrimLocation++;
						}
					}
					break;
				}
			}
			if (endTrim == 1) {
				break;
			}
		}
		if (lastTrimLocation > 32768) {
			fileSize = lastTrimLocation;
		}
		fseek(romFile, 0, SEEK_SET);
		
		
		// Set to 0 bank
		gba_flash_write_address_byte(0x1332, 0xAA);
		gba_flash_write_address_byte(0x1BBA, 0x55);
		gba_flash_write_address_byte(0x1332, 0); 
		
		// Read FRAM banks
		uint8_t framBanks[65];
		currAddr = 0x20000 / 2;
		set_number(currAddr, SET_START_ADDRESS);
		delay_ms(5);
		set_mode(GBA_READ_ROM);
		com_read_bytes(READ_BUFFER, 64);
		memcpy(&framBanks[0], readBuffer, 64);
		com_read_stop();
		
		delay_ms(5);
		set_number(0, SET_START_ADDRESS);
		
		
		// List games
		uint8_t gamesFound = 0;
		printf("\n           --Game--            --FRAM bank--\n");
		for (uint8_t x = 1; x < 64; x++) {
			gba_flash_write_address_byte(0x1332, 0xAA);
			gba_flash_write_address_byte(0x1BBA, 0x55);
			gba_flash_write_address_byte(0x1332, x); 
			
			if (gba_read_gametitle()) {
				printf ("Bank %i:", x);
				for (uint8_t t = 0; t < 20; t++) {
					printf("%c", gameTitle[t]);
				}
				
				if (x < 9) {
					printf("         %X", framBanks[x-1]);
				}
				else {
					printf("        %X", framBanks[x-1]);
				}
				
				printf("\n");
				gamesFound++;
			}
		}
		
		if (gamesFound == 0) {
			printf("No games found\n");
		}
		
		// Calculate file banks
		uint8_t fileBanks = originalFileSize / (1048576 * 2);
		if (fileBanks == 0) {
			fileBanks = 1;
		}
		
		// Free space
		uint8_t blocksBuffer[64];
		for (uint8_t x = 0; x < 64; x++) {
			blocksBuffer[x] = 0xFF;
		}
		
		// Read blocks file
		FILE *blocksFile = fopen("blocks.txt", "rb");
		if (blocksFile != NULL) {
			fread(blocksBuffer, 1, 64, blocksFile);
			fclose(blocksFile);
			
			printf("\nSpace utilisation: \n[");
			for (uint8_t x = 1; x < 64; x++) {
				if (blocksBuffer[x] == 0xFF) {
					printf("_");
				}
				else {
					printf("#");
				}
			}
			printf("]");
		}
		
		
		printf("\n\nROM file: %s (%i banks)\n", filenameOnly, fileBanks);
		
		
		// Read existing bank file
		FILE *bankReadFile = fopen("bank.txt", "rb");
		if (bankReadFile != NULL) {
			uint8_t bankRead[1];
			fread(bankRead, 1, 1, bankReadFile);
			fclose(bankReadFile);
			printf("\nThe next bank according to your last game written should be: %i\n", bankRead[0]);
		}
		printf("Which bank would you like to write the ROM file to? (1-63)\n");
		printf(">");
		
		
		// Read bank number
		char optionString[5];
		fgets(optionString, 5, stdin);
		int bankSelected = atoi(optionString);
		
		// Check if it will overwrite the loader
		if (bankSelected+fileBanks > 64) {
			printf("\n\nThe game is too large, it would overwrite the loader. \nPlease try again and select a lower slot number.\n");
			read_one_letter();
			return 1;
		}
		
		if (bankSelected+fileBanks < 63) {
			printf("\nThe next bank to write another game should be: %i\n", bankSelected+fileBanks);
			
			// Default blocks
			uint8_t blocksBuffer[64];
			for (uint8_t x = 0; x < 64; x++) {
				blocksBuffer[x] = 0xFF;
			}
			
			// Read blocks file
			FILE *blocksFile = fopen("blocks.txt", "rb");
			if (blocksFile != NULL) {
				fread(blocksBuffer, 1, 64, blocksFile);
				fclose(blocksFile);
			}
			
			// Check if bank selected will overwrite an existing game
			uint8_t overwritingGame = 0;
			uint8_t overwritingBlocks[64];
			uint8_t lastoverwriteBlock = 0;
			for (uint8_t x = bankSelected; x < bankSelected+fileBanks; x++) {
				if (blocksBuffer[x] != 0xFF) {
					overwritingGame = 1;
					overwritingBlocks[x] = 1;
					lastoverwriteBlock = x;
				}
				blocksBuffer[x] = 0; // Update blocks
			}
			if (overwritingGame == 1) {
				printf("\n*** The game will overwite an existing game (blocks ");
				for (uint8_t x = 0; x < 64; x++) {
					if (overwritingBlocks[x] == 1 && lastoverwriteBlock != x) {
						printf("%i,", x);
					}
					else if (overwritingBlocks[x] == 1 && lastoverwriteBlock == x) {
						printf("%i", x);
					}
				}
				
				printf(") *** \nPress enter to proceed.\n");
				read_one_letter();
			}
			
			// Write the updated blocks file
			blocksFile = fopen("blocks.txt", "wb");
			fwrite(blocksBuffer, 1, 64, blocksFile);
			fclose(blocksFile);
			
			// Save the next bank
			FILE *bankFile = fopen("bank.txt", "wb");
			uint8_t textBuffer[1];
			textBuffer[0] = bankSelected+fileBanks;
			fwrite(textBuffer, 1, 1, bankFile);
			fclose(bankFile);
		}
		
		// Change bank
		gba_flash_write_address_byte(0x1332, 0xAA);
		gba_flash_write_address_byte(0x1BBA, 0x55);
		gba_flash_write_address_byte(0x1332, bankSelected);
		
		
		
		if (gbxcartFirmwareVersion <= 19) {
			printf("Firmware R20 or higher is required for this functionality.\n");
			read_one_letter();
			return 1;
		}
		
		// Read rom a tiny bit before writing
		currAddr = 0x0000;	
		set_number(currAddr, SET_START_ADDRESS);
		delay_ms(5);
		set_mode(GBA_READ_ROM);
		delay_ms(5);
		com_read_bytes(READ_BUFFER, 64);
		com_read_stop();
		
		// Set end address as file size
		endAddr = fileSize;
		uint32_t endAddrAligned = fileSize;
		while ((endAddrAligned / 256) % 64 != 0) { // Align to 64 for printing progress
			endAddrAligned--;
		}
		
		// Verify chip ID
		gba_flash_write_address_byte(0xAAA, 0xAA);
		gba_flash_write_address_byte(0x555, 0x55);
		gba_flash_write_address_byte(0xAAA, 0x90);
		
		currAddr = 0x0000;	
		set_number(currAddr, SET_START_ADDRESS);
		delay_ms(5);
		set_mode(GBA_READ_ROM);
		delay_ms(5);
		com_read_bytes(READ_BUFFER, 64);
		com_read_stop();
		gba_flash_write_address_byte(0x000, 0xF0);
		
		printf("\nFlash ID: 0x%X,0x%X,0x%X,0x%X\n", readBuffer[0], readBuffer[1], readBuffer[2], readBuffer[3]);
		if (readBuffer[0] != 0x89 || readBuffer[1] != 0x0 || readBuffer[2] != 0x7E || readBuffer[3] != 0x22) { 
			printf("\n\nChip ID doesn't match 0x89 0x0 0x7E 0x22. Please re-seat the cartridge or press enter to continue anyway.\n");
			read_one_letter();
		}
		
		// Which write mode, buffered 32 bytes or 256 bytes at once
		uint8_t writeMode = GBA_FLASH_WRITE_BUFFERED_256BYTE; // 32 bytes
		if (readBuffer[0] == 0x89 && gbxcartFirmwareVersion >= 29) {
			writeMode = GBA_FLASH_WRITE_BUFFERED_128_WORDS; // 256 bytes
		}
		
		
		xmas_setup(endAddr / 28);
		
		printf("\nWriting to ROM (Flash cart) from %s\n", filenameOnly);
		printf("[             25%%             50%%             75%%            100%%]\n[");
		
		// Write ROM
		currAddr = 0x0000;	
		set_number(currAddr, SET_START_ADDRESS);
		delay_ms(5);
		while (currAddr < endAddr) {
			// Sector erase only performed for under 16MB files
			if (currAddr % 0x20000 == 0) { // Erase next sector
				gba_flash_write_address_byte(0xAAA, 0xAA);
				gba_flash_write_address_byte(0x555, 0x55);
				gba_flash_write_address_byte(0xAAA, 0x80);
				gba_flash_write_address_byte(0xAAA, 0xAA);
				gba_flash_write_address_byte(0x555, 0x55);
				gba_flash_write_address_byte((uint32_t) sector << 17, 0x30);
				sector++;
				
				//printf("Addr %i\n", currAddr);
				
				// Wait for first 2 bytes to be 0xFF
				wait_for_gba_flash_sector_ff(currAddr, 0xFF, 0xFF);
				
				set_number(currAddr / 2, SET_START_ADDRESS); // Divide address by 2
				delay_ms(5);
			}
			
			if (com_write_bytes_from_file_skip_FFs(writeMode, romFile, 256)) {
				com_wait_for_ack();
				currAddr += 256;
				readBytes += 256;
			}
			else {
				currAddr += 256;
				readBytes += 256;
				set_number(currAddr / 2, SET_START_ADDRESS); // Divide address by 2
				delay_ms(2);
			}
			
			print_progress_percent_addr(readBytes, endAddrAligned / 64);
			led_progress_percent(readBytes, endAddrAligned / 28);
		}
		
		gbx_cart_power_down();
		
		printf("\n\nCompleted");
		read_one_letter();
	}
	else {
		read_config();
		
		// Open COM port
		if (com_test_port() == 0) {
			printf("Device not connected and couldn't be auto detected. Please make sure the COM port isn't open elsewhere. If you are using Linux/Mac, make sure you use \"sudo\".\n");
			read_one_letter();
			return 1;
		}
		printf("Connected on COM port: %i at ", cport_nr+1);
		
		// Break out of any existing functions on ATmega
		set_mode('0');
		RS232_flushRX(cport_nr);
		
		// Get cartridge mode - Gameboy or Gameboy Advance
		cartridgeMode = request_value(CART_MODE);
		
		// Get PCB version
		gbxcartPcbVersion = request_value(READ_PCB_VERSION);
		xmas_wake_up();
		
		// Increase to 1.7Mbit baud for v1.4 PCB
		#ifdef _WIN32
		if (gbxcartPcbVersion == PCB_1_4) {
			set_mode(USART_1_7M_SPEED);
			RS232_CloseComport(cport_nr);
			delay_ms(200);
			
			if (RS232_OpenComport(cport_nr, 1700000, "8N1") == 0) {
				printf("1.7M Baud\n");
			}
		}
		else {
			printf("1M Baud\n");
		}
		#else
			printf("1M Baud\n");
		#endif
		
		char pcbVersionString[20];
		if (gbxcartPcbVersion == PCB_1_0) {
			strncpy(pcbVersionString, "v1.0", 5);
		}
		else if (gbxcartPcbVersion == PCB_1_1) {
			strncpy(pcbVersionString, "v1.1/v1.2", 10);
		}
		else if (gbxcartPcbVersion == PCB_1_3) {
			strncpy(pcbVersionString, "v1.3", 5);
		}
		else if (gbxcartPcbVersion == PCB_1_4) {
			strncpy(pcbVersionString, "v1.4", 5);
		}
		else if (gbxcartPcbVersion == MINI) {
			strncpy(pcbVersionString, "Mini", 5);
		}
		else if (gbxcartPcbVersion == GBXMAS) {
			strncpy(pcbVersionString, "XMAS", 5);
		}
		
		// Get firmware version
		gbxcartFirmwareVersion = request_value(READ_FIRMWARE_VERSION);
		printf("PCB version: %s, Firmware version: %i\n", pcbVersionString, gbxcartFirmwareVersion);
		
		
		// Power up to GBA mode
		set_mode(VOLTAGE_3_3V);
		
		// GBx v1.4 - Power up the cart if not already powered up and flush buffer
		gbx_cart_power_up();
		RS232_flushRX(cport_nr);
		
		if (gbxcartPcbVersion == GBXMAS) {
			set_mode('!');
			delay_ms(50); // Wait for ATmega169 to WDT reset if more than 2 mins idle
		}
		
		// Get cartridge mode - Gameboy or Gameboy Advance
		cartridgeMode = request_value(CART_MODE);
		
		
		if (gbxcartPcbVersion == PCB_1_0) {
			printf("\nPCB v1.0 is not supported for this function.");
			read_one_letter();
			return 1;
		}
		
		if (gbxcartFirmwareVersion <= 8) {
			printf("Firmware R9 or higher is required for this functionality.\n");
			read_one_letter();
			return 1;
		}
		
		
		uint8_t inMenu = 1;
		while (inMenu == 1) {
		
			// Set to 0 bank
			gba_flash_write_address_byte(0x1332, 0xAA);
			gba_flash_write_address_byte(0x1BBA, 0x55);
			gba_flash_write_address_byte(0x1332, 0); 
			
			// Read FRAM banks
			uint8_t framBanks[65];
			currAddr = 0x20000 / 2;
			set_number(currAddr, SET_START_ADDRESS);
			delay_ms(5);
			set_mode(GBA_READ_ROM);
			com_read_bytes(READ_BUFFER, 64);
			memcpy(&framBanks[0], readBuffer, 64);
			com_read_stop();
			
			delay_ms(5);
			set_number(0, SET_START_ADDRESS);
			
			
			// List games
			uint8_t gamesFound = 0;
			printf("\n           --Game--            --FRAM bank--\n");
			for (uint8_t x = 1; x < 64; x++) {
				gba_flash_write_address_byte(0x1332, 0xAA);
				gba_flash_write_address_byte(0x1BBA, 0x55);
				gba_flash_write_address_byte(0x1332, x); 
				
				if (gba_read_gametitle()) {
					printf ("Bank %i:", x);
					for (uint8_t t = 0; t < 20; t++) {
						printf("%c", gameTitle[t]);
					}
					
					if (x < 9) {
						printf("         %X", framBanks[x-1]);
					}
					else {
						printf("        %X", framBanks[x-1]);
					}
					
					printf("\n");
					gamesFound++;
				}
			}
			
			if (gamesFound == 0) {
				printf("No games found\n");
			}
			
			
			// Set to 0 bank
			gba_flash_write_address_byte(0x1332, 0xAA);
			gba_flash_write_address_byte(0x1BBA, 0x55);
			gba_flash_write_address_byte(0x1332, 0);
			
			
			
			printf("\nWould you like to change to a game [g], change a game's FRAM bank [b] or change a game's name[n]?\n");
			printf(">");
			
			// Read number
			char optionSelected = read_one_letter();
			
			if (optionSelected == 'g') {
				printf("\nWhich game would you like to change to? (1-63)\n");
				printf(">");
				
				// Read bank number
				char optionString[5];
				fgets(optionString, 5, stdin);
				int gameSelected = atoi(optionString);
				
				// Change bank
				gba_flash_write_address_byte(0x1332, 0xAA);
				gba_flash_write_address_byte(0x1BBA, 0x55);
				gba_flash_write_address_byte(0x1332, (framBanks[gameSelected-1] << 6) | gameSelected);
				//printf("0x%X\n", (framBanks[gameSelected-1] << 6) | gameSelected);
				
				printf("\nCompleted. Please close this program and use the GUI to read/write the save.\n");
				read_one_letter();
			}		
			else if (optionSelected == 'b') {
				printf("\nWhich game's FRAM bank would you like to change?\n");
				printf(">");
				
				// Read number
				char optionString[5];
				fgets(optionString, 5, stdin);
				int gameSelected = atoi(optionString) - 1;
				
				printf("\nAnd to which FRAM bank? (0-3)\n");
				printf(">");
				
				// Read number
				fgets(optionString, 5, stdin);
				int framBankSelected = atoi(optionString);
				
				// Change
				framBanks[gameSelected] = framBankSelected;
				
				
				// Verify chip ID
				gba_flash_write_address_byte(0xAAA, 0xAA);
				gba_flash_write_address_byte(0x555, 0x55);
				gba_flash_write_address_byte(0xAAA, 0x90);
				
				currAddr = 0x0000;	
				set_number(currAddr, SET_START_ADDRESS);
				delay_ms(5);
				set_mode(GBA_READ_ROM);
				delay_ms(5);
				com_read_bytes(READ_BUFFER, 64);
				com_read_stop();
				gba_flash_write_address_byte(0x000, 0xF0);
				
				printf("\nFlash ID: 0x%X,0x%X,0x%X,0x%X\n", readBuffer[0], readBuffer[1], readBuffer[2], readBuffer[3]);
				if (readBuffer[0] != 0x89 || readBuffer[1] != 0x0 || readBuffer[2] != 0x7E || readBuffer[3] != 0x22) { 
					printf("\n\nChip ID doesn't match 0x89 0x0 0x7E 0x22. Please re-seat the cartridge or press enter to continue anyway.\n");
					read_one_letter();
				}
				printf("Please wait... ");
				
				// Write banks
				currAddr = 0x20000;
				set_number(currAddr / 2, SET_START_ADDRESS);
				delay_ms(5);
				
				// Sector erase
				uint8_t sector = 1;
				gba_flash_write_address_byte(0xAAA, 0xAA);
				gba_flash_write_address_byte(0x555, 0x55);
				gba_flash_write_address_byte(0xAAA, 0x80);
				gba_flash_write_address_byte(0xAAA, 0xAA);
				gba_flash_write_address_byte(0x555, 0x55);
				gba_flash_write_address_byte((uint32_t) sector << 17, 0x30);
				
				// Wait for first 2 bytes to be 0xFF
				wait_for_gba_flash_sector_ff(currAddr, 0xFF, 0xFF);
				
				set_number(currAddr / 2, SET_START_ADDRESS); // Divide address by 2
				delay_ms(5);
				
				// Write
				for (uint8_t x = 0; x < 64; x += 2) {
					gba_flash_write_address_byte(0xAAA, 0xAA);
					gba_flash_write_address_byte(0x555, 0x55);
					gba_flash_write_address_byte(0xAAA, 0xA0);
					gba_flash_write_address_byte(0x20000 + x, ((framBanks[x+1] << 8) | framBanks[x]));
					//printf("%i, 0x%X\n", x, (framBanks[x+1] << 8) | framBanks[x]);
				}
				
				printf("Completed\n\n");
			}
			else if (optionSelected == 'n') {
				printf("\nWhich game's name would you like to change?\n");
				printf(">");
				
				// Read bank number
				char optionString[5];
				fgets(optionString, 5, stdin);
				int gameSelected = atoi(optionString);
				
				// Change bank
				gba_flash_write_address_byte(0x1332, 0xAA);
				gba_flash_write_address_byte(0x1BBA, 0x55);
				gba_flash_write_address_byte(0x1332, (framBanks[gameSelected-1] << 6) | gameSelected);
				
				
				gba_read_gametitle();
				for (uint8_t t = 0; t < 20; t++) {
					printf("%c", gameTitle[t]);
				}
				printf("\n");
				
				
				printf("\nPlease type a new game title for it (max 12 characters)\n");
				printf(">");
				
				char titleString[13];
				fgets(titleString, 13, stdin);
				
				
				// Read first 0x20000
				currAddr = 0x0000;
				endAddr = 0x20000;
				set_number(currAddr, SET_START_ADDRESS);
				set_mode(GBA_READ_ROM);
				
				uint8_t romBuffer[0x20000];
				while (currAddr < endAddr) {
					com_read_bytes(READ_BUFFER, 64);
					memcpy(&romBuffer[currAddr], readBuffer, 64);
					currAddr += 64;
					
					if (currAddr < endAddr) {
						com_read_cont();
					}
				}
				com_read_stop();
				
				
				// Change the title
				uint8_t z = 0;
				for (uint16_t x = 0xA0; x < 0xA0 + 12; x++) {
					if (titleString[z] != 0) {
						romBuffer[x] = titleString[z];
					}
					else {
						romBuffer[x] = 0;
					}
					z++;
				}
				
				
				// Verify chip ID
				gba_flash_write_address_byte(0xAAA, 0xAA);
				gba_flash_write_address_byte(0x555, 0x55);
				gba_flash_write_address_byte(0xAAA, 0x90);
				
				currAddr = 0x0000;	
				set_number(currAddr, SET_START_ADDRESS);
				delay_ms(5);
				set_mode(GBA_READ_ROM);
				delay_ms(5);
				com_read_bytes(READ_BUFFER, 64);
				com_read_stop();
				gba_flash_write_address_byte(0x000, 0xF0);
				
				printf("\nFlash ID: 0x%X,0x%X,0x%X,0x%X\n", readBuffer[0], readBuffer[1], readBuffer[2], readBuffer[3]);
				if (readBuffer[0] != 0x89 || readBuffer[1] != 0x0 || readBuffer[2] != 0x7E || readBuffer[3] != 0x22) { 
					printf("\n\nChip ID doesn't match 0x89 0x0 0x7E 0x22. Please re-seat the cartridge or press enter to continue anyway.\n");
					read_one_letter();
				}
				
				// Which write mode, buffered 32 bytes or 256 bytes at once
				uint8_t writeMode = GBA_FLASH_WRITE_BUFFERED_256BYTE; // 32 bytes
				if (readBuffer[0] == 0x89 && gbxcartFirmwareVersion >= 29) {
					writeMode = GBA_FLASH_WRITE_BUFFERED_128_WORDS; // 256 bytes
				}
				
				printf("Please wait... ");
				
				
				// Write changes
				currAddr = 0x0000;
				endAddr = 0x20000;
				set_number(currAddr, SET_START_ADDRESS);
				delay_ms(5);
				uint8_t sector = 0;
				while (currAddr < endAddr) {
					if (currAddr % 0x20000 == 0) { // Erase next sector
						gba_flash_write_address_byte(0xAAA, 0xAA);
						gba_flash_write_address_byte(0x555, 0x55);
						gba_flash_write_address_byte(0xAAA, 0x80);
						gba_flash_write_address_byte(0xAAA, 0xAA);
						gba_flash_write_address_byte(0x555, 0x55);
						gba_flash_write_address_byte((uint32_t) sector << 17, 0x30);
						sector++;
						
						// Wait for first 2 bytes to be 0xFF
						wait_for_gba_flash_sector_ff(currAddr, 0xFF, 0xFF);
						
						set_number(currAddr / 2, SET_START_ADDRESS); // Divide address by 2
						delay_ms(5);
					}
					
					// Load write buffer
					for (uint16_t x = 0; x < 256; x++) {
						writeBuffer[x] = romBuffer[currAddr+x];
					}
					com_write_bytes_from_file(writeMode, NULL, 256);
					com_wait_for_ack();
					currAddr += 256;
				}
				
				printf("Completed\n\n");
			}
			else {
				printf("No option selected\n\n");
			}
		}
	}
	
	printf("\n");
	xmas_idle_on();
	
	return 0;
}
